@@ -38,6 +38,7 @@
 #include <errno.h>
 
 #include "glib.h"
+#include "gstdio.h"
 
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -318,6 +319,119 @@ find_first_that(gpointer key,
 }
 
 
+static gboolean
+test_g_mkdir_with_parents_1 (const gchar *base)
+{
+  char *p0 = g_build_filename (base, "fum", NULL);
+  char *p1 = g_build_filename (p0, "tem", NULL);
+  char *p2 = g_build_filename (p1, "zap", NULL);
+  FILE *f;
+
+  g_remove (p2);
+  g_remove (p1);
+  g_remove (p0);
+
+  if (g_file_test (p0, G_FILE_TEST_EXISTS))
+    {
+      g_print ("failed, %s exists, cannot test g_mkdir_with_parents\n", p0);
+      return FALSE;
+    }
+
+  if (g_file_test (p1, G_FILE_TEST_EXISTS))
+    {
+      g_print ("failed, %s exists, cannot test g_mkdir_with_parents\n", p1);
+      return FALSE;
+    }
+
+  if (g_file_test (p2, G_FILE_TEST_EXISTS))
+    {
+      g_print ("failed, %s exists, cannot test g_mkdir_with_parents\n", p2);
+      return FALSE;
+    }
+
+  if (g_mkdir_with_parents (p2, 0666) == -1)
+    {
+      g_print ("failed, g_mkdir_with_parents(%s) failed: %s\n", p2, g_strerror (errno));
+      return FALSE;
+    }
+
+  if (!g_file_test (p2, G_FILE_TEST_IS_DIR))
+    {
+      g_print ("failed, g_mkdir_with_parents(%s) succeeded, but %s is not a directory\n", p2, p2);
+      return FALSE;
+    }
+
+  if (!g_file_test (p1, G_FILE_TEST_IS_DIR))
+    {
+      g_print ("failed, g_mkdir_with_parents(%s) succeeded, but %s is not a directory\n", p2, p1);
+      return FALSE;
+    }
+
+  if (!g_file_test (p0, G_FILE_TEST_IS_DIR))
+    {
+      g_print ("failed, g_mkdir_with_parents(%s) succeeded, but %s is not a directory\n", p2, p0);
+      return FALSE;
+    }
+
+  g_rmdir (p2);
+  if (g_file_test (p2, G_FILE_TEST_EXISTS))
+    {
+      g_print ("failed, did g_rmdir(%s), but %s is still there\n", p2, p2);
+      return FALSE;
+    }
+
+  g_rmdir (p1);
+  if (g_file_test (p1, G_FILE_TEST_EXISTS))
+    {
+      g_print ("failed, did g_rmdir(%s), but %s is still there\n", p1, p1);
+      return FALSE;
+    }
+
+  f = g_fopen (p1, "w");
+  if (f == NULL)
+    {
+      g_print ("failed, couldn't create file %s\n", p1);
+      return FALSE;
+    }
+  fclose (f);
+  
+  if (g_mkdir_with_parents (p1, 0666) == 0)
+    {
+      g_print ("failed, g_mkdir_with_parents(%s) succeeded, even if %s is a file\n", p1, p1);
+      return FALSE;
+    }
+
+  if (g_mkdir_with_parents (p2, 0666) == 0)
+    {
+      g_print ("failed, g_mkdir_with_parents(%s) succeeded, even if %s is a file\n", p2, p1);
+      return FALSE;
+    }
+
+  g_remove (p2);
+  g_remove (p1);
+  g_remove (p0);
+
+  return TRUE;
+}
+
+static gboolean
+test_g_mkdir_with_parents (void)
+{
+  g_print ("checking g_mkdir_with_parents()...");
+  if (!test_g_mkdir_with_parents_1 ("hum"))
+    return FALSE;
+  g_remove ("hum");
+  if (!test_g_mkdir_with_parents_1 ("hii///haa/hee"))
+    return FALSE;
+  g_remove ("hii/haa/hee");
+  g_remove ("hii/haa");
+  g_remove ("hii");
+  if (!test_g_mkdir_with_parents_1 (g_get_current_dir ()))
+    return FALSE;
+
+  return TRUE;
+}
+
 int
 main (int   argc,
       char *argv[])
@@ -330,6 +444,8 @@ main (int   argc,
   GMemChunk *mem_chunk;
   GStringChunk *string_chunk;
   GTimer *timer, *timer2;
+  gdouble elapsed;
+  gulong elapsed_usecs;
   gint nums[10] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9 };
   gint morenums[10] = { 8, 9, 7, 0, 3, 2, 5, 1, 4, 6};
   gchar *string;
@@ -436,8 +552,7 @@ main (int   argc,
   g_free (string);
   g_print ("user: %s\n", g_get_user_name ());
   g_print ("real: %s\n", g_get_real_name ());
-  s = g_get_home_dir ();
-  g_print ("home: %s\n", s ? s : "NULL!");
+  g_print ("host: %s\n", g_get_host_name ());
   s = g_get_home_dir ();
   g_print ("home: %s\n", s ? s : "NULL!");
   s = g_get_user_data_dir ();
@@ -481,7 +596,6 @@ main (int   argc,
   string = g_path_get_basename ("/foo/file");
   g_assert (strcmp (string, "file") == 0);
   g_free (string);
-  g_print ("ok\n");
 #endif
 
   g_print ("checking g_path_get_dirname()...");
@@ -525,6 +639,9 @@ main (int   argc,
   if (n_skip_root_checks)
     g_print ("ok\n");
 
+  if (test_g_mkdir_with_parents ())
+    g_print ("ok\n");
+
   g_print ("checking doubly linked lists...");
 
   list = NULL;
@@ -1013,16 +1130,20 @@ main (int   argc,
   g_usleep((29*G_USEC_PER_SEC)/5); /* run timer for 5.8 seconds */
   g_timer_stop(timer);
 
-  g_print ("\t=> total elapsed = %.2f seconds (should be: 9.00 seconds)\n\n", g_timer_elapsed(timer, NULL));
+  elapsed = g_timer_elapsed (timer, &elapsed_usecs);
+  g_print ("\t=> timer = %.6f = %d.%06ld (should be: 9.000000) (%.6f off)\n", elapsed, (int) elapsed, elapsed_usecs, ABS (elapsed - 9.));
 
-  if (g_timer_elapsed(timer, NULL) > 8.8 && g_timer_elapsed(timer, NULL) < 9.2)
+  if (elapsed > 8.8 && elapsed < 9.2)
     g_print ("g_timer_continue ... ok\n\n");
   else
     g_print ("g_timer_continue ... ***** FAILED *****\n\n");
 
   g_timer_stop(timer2);
 
-  if (g_timer_elapsed(timer2, NULL) > (8.8+6.5) && g_timer_elapsed(timer2, NULL) < (9.2+6.5))
+  elapsed = g_timer_elapsed(timer2, &elapsed_usecs);
+  g_print ("\t=> timer2 = %.6f = %d.%06ld (should be: %.6f) (%.6f off)\n\n", elapsed, (int) elapsed, elapsed_usecs, 9.+6.5, ABS (elapsed - (9.+6.5)));
+
+  if (elapsed > (8.8+6.5) && elapsed < (9.2+6.5))
     g_print ("timer2 ... ok\n\n");
   else
     g_print ("timer2 ... ***** FAILED *****\n\n");
@@ -1286,6 +1407,8 @@ main (int   argc,
   g_print ("found more.com as %s\n", g_find_program_in_path ("more.com"));
   g_print ("found regedit as %s\n", g_find_program_in_path ("regedit"));
 
+  g_print ("a Win32 error message: %s\n", g_win32_error_message (2));
+
 #endif
 
   g_print ("checking file functions...\n");
